from typing import Optional, Any, Dict, Iterable, Mapping
import ast
import os
import inspect
from RestrictedPython import compile_restricted
from RestrictedPython.Guards import safe_builtins
from RestrictedPython.Eval import default_guarded_getiter
from RestrictedPython.Guards import guarded_iter_unpack_sequence, safer_getattr
from typing import Mapping, Sequence
from copy import copy


global_dict = {
    '__name__': '__main__',
    '__metaclass__': type,
    '_getiter_': default_guarded_getiter,
    '_iter_unpack_sequence_': guarded_iter_unpack_sequence,
    '_getattr_': safer_getattr,
    '__builtins__': safe_builtins,
}


# def restricted_exec(
#     source,
#     globals: Optional[Dict[str, Any]] = None,
#     locals: Optional[Mapping[str, object]] = None,
#     admission: Optional[Iterable[object]] = [],
#     avoidance: Optional[Iterable[object]] = None
# ):
#     compiled_code = compile_restricted(source, '<string>', 'exec')
#     if admission is not None:
#         def restricted_import(
#             name: str,
#             globals: Optional[Dict[str, Any]] = None,
#             locals: Optional[Mapping[str, object]] = None,
#             fromlist: Sequence[str] = (),
#             level: int = 0,
#         ):
#             imported = __import__(name, globals, locals, fromlist, level)
#             can_be_imported = True
#             for obj in admission:
#                 can_be_imported = can_be_imported and ((obj.__name__ in name and obj is imported) or (obj.__name__ in fromlist and hasattr(imported, obj.__name__)))
#             length = len(fromlist)
#             if length == 0:
#                 raise ImportError(f"'{name}' is not allowed to be imported in this context.")
#             else:
#                 fromstr = f"'{fromlist[0]}'"
#                 if length > 1:
#                     for i, frompart in enumerate(fromlist[1:]):
#                         if i == length - 2:
#                             fromstr += f" and '{frompart}'"
#                         else:
#                             fromstr += f", '{frompart}'"
#                 raise ImportError(f"{fromstr} in the module '{name}' is not allowed to be imported in this context.")
#         import_func = restricted_import
#     else:
#         import_func = __import__
#     if avoidance is not None:
#         def restricted_import(
#             name: str,
#             globals: Optional[Dict[str, Any]] = None,
#             locals: Optional[Mapping[str, object]] = None,
#             fromlist: Sequence[str] = (),
#             level: int = 0,
#         ):
#             imported = __import__(name, globals, locals, fromlist, level)
#             for obj in admission:
#                 if (obj.__name__ in name and obj is imported) or (obj.__name__ in fromlist and hasattr(imported, obj.__name__)):
#                     raise 
#     exec_global_dict = copy(global_dict)
#     exec_global_dict.update({})
#     exec(compiled_code, globals, locals)


def restricted_eval(source, globals: Optional[Dict[str, Any]] = None, locals: Optional[Mapping[str, object]] = None):
    compiled_code = compile_restricted(source, '<string>', 'eval')
    if globals is not None:
        globals = copy(globals)
        globals.update(global_dict)
    return eval(compiled_code, globals, locals)


def check_avoidance(variable, avoidance, name = ''):
    for avoiding_variable in avoidance:
        is_in_avoidance = (variable is getattr(avoiding_variable, variable.__name__, None))
        if (
            (inspect.ismodule(avoiding_variable) and is_in_avoidance) or
            (isinstance(avoiding_variable, type) and (issubclass(variable, avoiding_variable) or isinstance(variable, avoiding_variable) or is_in_avoidance)) or
            (variable is avoiding_variable)
        ):
            raise RuntimeError(f"'{name} ({variable})' is forbidden in this context")


class AvoidanceVisitor(ast.NodeVisitor):
    def __init__(self, avoidance = [], globals = None, locals = None):
        super().__init__()
        self.avoidance = avoidance
        self.globals = globals
        self.locals = locals
    
    def visit_Name(self, node: ast.Name):
        name = node.id
        variable = eval(name, self.globals, self.locals)
        check_avoidance(variable, self.avoidance, name)
        self.generic_visit(node)


def avoidance_eval(
    source: str,
    globals: Optional[Dict[str, Any]] = None,
    locals: Optional[Mapping[str, object]] = None,
    avoidance: Iterable[Any] = [os]
):
    tree = ast.parse(source, mode='eval')
    last_globals = inspect.currentframe().f_back.f_globals
    last_locals = inspect.currentframe().f_back.f_locals
    if globals is None:
        globals = {}
    if locals is None:
        locals = {}
    globals.update(last_globals)
    locals.update(last_locals)
    visitor = AvoidanceVisitor(avoidance, globals, locals)
    visitor.visit(tree)
    return restricted_eval(source, globals, locals)


from typing import Dict
import os
import inspect
import yaml
from numba import jit
from .utils.restricted_parse import avoidance_eval

@jit(nopython=True)
def find_mark(buffer: str):
    char_list = [':', ',', ']', '}', ')']
    for i, char in enumerate(buffer):
        if char in char_list:
            return i
    return -1

def end_mark(node: yaml.ScalarNode):
    return node.end_mark.buffer[node.end_mark.index]

def is_real_str_value(node: yaml.ScalarNode):
    if node.style is None:
        buffer = node.end_mark.buffer[node.start_mark.index:]
        mark_index = find_mark(buffer)
        if mark_index != -1:
            substring_before_mark = buffer[:mark_index]
            if not '!!str' in substring_before_mark:
                return False
    return True

def construct_yaml_str(loader: yaml.Loader, node: yaml.ScalarNode, globals = None, locals = None, avoidance = [os]):
    value = loader.construct_scalar(node)
    if end_mark(node) != ':':
        if not is_real_str_value(node):
            value = avoidance_eval(value, globals, locals, avoidance)
    return value

# def construct_yaml_str(loader: yaml.Loader, node: yaml.ScalarNode):
#     value = loader.construct_scalar(node)
#     if node.end_mark.buffer[node.end_mark.index] != ':':
#         if node.style is None:
#             buffer = node.end_mark.buffer[node.start_mark.index:]
#             mark_index = find_mark(buffer)
#             if mark_index != -1:
#                 substring_before_mark = buffer[:mark_index]
#                 if not '!!str' in substring_before_mark:
#                     value = safeeval(value)
#     return value

def construct_yaml_seq(loader: yaml.Loader, node: yaml.SequenceNode):
    data = []
    yield data
    data.extend(construct_sequence(loader, node))

def construct_sequence(loader: yaml.Loader, node: yaml.SequenceNode, deep=False):
    if not isinstance(node, yaml.SequenceNode):
        raise yaml.constructor.ConstructorError(None, None,
                "expected a sequence node, but found %s" % node.id,
                node.start_mark)
    return [construct_object(loader, child, deep=deep) for child in node.value]

def construct_object(loader: yaml.Loader, node: yaml.SequenceNode, deep=False):
    string_tag = 'tag:yaml.org,2002:str'
    if node.tag == string_tag:
        return loader.__class__.yaml_constructors[string_tag](loader, node)
    else:
        return loader.construct_object(node, deep=deep)

def back_locals() -> Dict:
    return inspect.currentframe().f_back.f_back.f_locals

# def import_target(path: str, need_target = True):
#     last_point_index = path.rfind('.')
#     if last_point_index == -1:
#         to_be_imported = path
#     else:
#         to_be_imported = path[last_point_index + 1:]
#         path = 'from ' + path[:last_point_index] + ' import ' + to_be_imported
#         safeeval(path)
#     if need_target:
#         return safeeval(to_be_imported)


import collections.abc, types
import os
import copy
import re
import yaml.composer
import yaml.parser
import yaml.scanner
from . import special_keys
from .namespace import Namespace
from .executable_construction import ExecutableConstructor
from .utils.restricted_parse import avoidance_eval, check_avoidance

class TypeOnlyError(Exception): ...

def extract_arguments(*args, **kwargs):
    return args, kwargs

def is_custom_type_node(node: yaml.Node):
    # If the type is not a custom type but a YAML type, its tag will
    # start with 'tag:yaml.org,2002:', even if the prefix '!!' is specified.
    return node.tag.startswith('!')

class InstantiableConstructor(ExecutableConstructor):
    '''
    This constructor not only executes codes based on `ExecutableConstructor` but
    also can instantiate objects when specifying the type after the mark `!`. All
    YAML attributed dictionary can also be regared as an object, which is a
    namespace in this constructor, which means they can be accessed by not only
    `[] (__getitem__)` but also dot operator (__getattr__), all not existed attribute
    will result in a `None` return.
    '''
    def __init__(
        self,
        avoidance = [os],
        args_id = special_keys.args_id,
        kwargs_id = special_keys.kwargs_id,
        run_id = special_keys.run_id,
        import_id = special_keys.import_id,
        imported_targets_id = special_keys.imported_targets_id
    ):
        super().__init__(avoidance, run_id, import_id, imported_targets_id)
        self.__args_id__ = args_id
        self.__kwargs_id__ = kwargs_id
    
    def construct_document(self, node):
        # Make the final top element `Namespace`.
        data = super().construct_document(node)
        if isinstance(data, dict):
            data = Namespace(data)
        return data
    
    def construct_value(self, parent_node, node, deep=False):
        if isinstance(parent_node, yaml.MappingNode) and node.tag == 'tag:yaml.org,2002:map' and (not node.flow_style):
            node.tag = 'tag:yaml.org,2002:namespace'
        return self.construct_object(node, deep)

    def construct_yaml_namespace(self, node):
        data = Namespace()
        yield data
        value = self.construct_mapping(node)
        data.update(value)
    
    def construct_yaml_instance(self, node: yaml.Node):
        instance_type: str = node.tag[1:]
        target_names = instance_type.split('.')
        target_name = target_names[0]
        target = self.__vars__.get(target_name)
        if target is None:
            target = self.import_target(instance_type)
        else:
            target_names = target_names[1:]
            for target_name in target_names:
                target = getattr(target, target_name)
        check_avoidance(target, self.avoidance, target_name)
        try:
            args, kwargs = self.construct_arguments(node)
            return target(*args, **kwargs)
        except TypeOnlyError:
            return target
    
    def escape_special_key(self, key, special_key):
        # Make the special key, like `__import_id__` key, as a key but remove the
        # special meanings.
        if key[:len(special_key) + 1] == f'\\{special_key}':
            key = key[1:]
        return key
    
    def check_key_format(self, key, key_node):
        # Ensure that the key is Pythonicly named.
        if re.fullmatch(r'[^\W\d]\w*', key) is None:
            raise yaml.constructor.ConstructorError(
                "The identifier of the key should comply with Python naming conventions.",
                key_node.start_mark)
    
    def construct_instantiating_mapping(self, node: yaml.MappingNode, deep=False):
        # Create mapping without delay.
        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if not isinstance(key, collections.abc.Hashable):
                raise yaml.constructor.ConstructorError("while constructing a mapping", node.start_mark,
                        "found unhashable key", key_node.start_mark)
            value = self.construct_instantiating_value(value_node, deep=deep)
            mapping[key] = value
        return mapping
    
    def construct_instantiating_sequence(self, node: yaml.SequenceNode, deep=False):
        # Create sequence without delay.
        return [self.construct_instantiating_value(child, deep=deep) for child in node.value]

    def construct_instantiating_value(self, node, deep=False):
        if is_custom_type_node(node):
            return self.construct_object(node, deep)
        else:
            # Ensure all variables will be initialized without using `yield`.
            if isinstance(node, yaml.MappingNode):
                return self.construct_instantiating_mapping(node, deep)
            elif isinstance(node, yaml.SequenceNode):
                return self.construct_instantiating_sequence(node, deep)
            else:
                return self.construct_object(node, deep)
    
    def construct_arguments(self, node: yaml.Node, deep=False):
        # Construct the arguments that is necessary to be passed into the `__init__`.
        args, kwargs = [], {}
        if isinstance(node, yaml.MappingNode):
            '''
            instance: !type
                __args__: [1, 2, 3]
                __kwargs__:
                    first: "1st"
                    second: '2nd'
            '''
            self.flatten_mapping(node)
            for key_node, value_node in node.value:
                key = self.construct_object(key_node, deep=deep)
                if not isinstance(key, collections.abc.Hashable):
                    raise yaml.constructor.ConstructorError("While constructing a mapping", node.start_mark,
                            "found unhashable key", key_node.start_mark)
                if key == self.__args_id__:
                    args += self.construct_instantiating_sequence(value_node, deep)
                elif key == self.__kwargs_id__:
                    kwargs.update(self.construct_instantiating_mapping(value_node, deep))
                else:
                    if isinstance(value_node, yaml.MappingNode):
                        raise yaml.constructor.ConstructorError("A `dict` passed as an argument should be wrapped in the `[]`.", value_node.end_mark)
                    pair = self.import_target_from_node(node, key_node, value_node, deep)
                    if pair is None:
                        key = self.escape_special_key(key, self.__import_id__)
                    else:
                        key, _ = pair
                    key = self.escape_special_key(key, self.__args_id__)
                    key = self.escape_special_key(key, self.__kwargs_id__)
                    self.check_key_format(key, key_node)
                    instantiating_value = self.construct_instantiating_value(value_node, deep=deep)
                    if key != self.__run_id__:
                        kwargs.update({key : instantiating_value})
        else:
            value_scalar = node.value
            value_node = copy.deepcopy(node)
            if isinstance(value_node, yaml.ScalarNode):
                if value_node.style is None:
                    if value_scalar.startswith('~(') and value_scalar.endswith(')'):
                        '''
                        instance: !type ~(1, 2, 3, first="1st", second='2nd')
                        '''
                        new_args, new_kwargs = avoidance_eval(f'{extract_arguments.__code__.co_name}{value_scalar[1:]}', avoidance=self.avoidance)
                        args += new_args
                        kwargs.update(new_kwargs)
                        return args, kwargs
                    elif value_scalar == '':
                        raise TypeOnlyError()
                    # else:
                    #     '''
                    #     instance: !type "the only arg"
                    #     '''
                    #     value = safeeval(value_scalar, avoidance=self.avoidance)
                else:
                    value = value_scalar
                args.append(value)
            # elif isinstance(value_node, yaml.MappingNode):
            #     '''
            #     instance: !type {'first': "1st", 'second': '2nd'}
            #     '''
            #     value = self.construct_mapping_directly(value_node, deep)
            elif isinstance(value_node, yaml.SequenceNode):
                '''
                instance: !type [1, 2, 3]
                '''
                values = self.construct_instantiating_sequence(value_node, deep)
                args += values
        return args, kwargs
    
    def construct_object(self, node, deep=False):
        if node in self.constructed_objects:
            return self.constructed_objects[node]
        if deep:
            old_deep = self.deep_construct
            self.deep_construct = True
        if node in self.recursive_objects:
            raise yaml.constructor.ConstructorError(None, None,
                    "found unconstructable recursive node", node.start_mark)
        self.recursive_objects[node] = None
        constructor = None
        tag_suffix = None
        if node.tag in self.yaml_constructors:
            constructor = self.yaml_constructors[node.tag]
        elif is_custom_type_node(node): # The only modification in this method.
            constructor = self.__class__.construct_yaml_instance
        else:
            for tag_prefix in self.yaml_multi_constructors:
                if tag_prefix is not None and node.tag.startswith(tag_prefix):
                    tag_suffix = node.tag[len(tag_prefix):]
                    constructor = self.yaml_multi_constructors[tag_prefix]
                    break
            else:
                if None in self.yaml_multi_constructors:
                    tag_suffix = node.tag
                    constructor = self.yaml_multi_constructors[None]
                elif None in self.yaml_constructors:
                    constructor = self.yaml_constructors[None]
                elif isinstance(node, yaml.ScalarNode):
                    constructor = self.__class__.construct_scalar
                elif isinstance(node, yaml.SequenceNode):
                    constructor = self.__class__.construct_sequence
                elif isinstance(node, yaml.MappingNode):
                    constructor = self.__class__.construct_mapping
        if tag_suffix is None:
            data = constructor(self, node)
        else:
            data = constructor(self, tag_suffix, node)
        if isinstance(data, types.GeneratorType):
            generator = data
            data = next(generator)
            if self.deep_construct:
                for dummy in generator:
                    pass
            else:
                self.state_generators.append(generator)
        self.constructed_objects[node] = data
        del self.recursive_objects[node]
        if deep:
            self.deep_construct = old_deep
        return data

InstantiableConstructor.add_constructor('tag:yaml.org,2002:namespace', InstantiableConstructor.construct_yaml_namespace)

class InstantiableLoader(yaml.reader.Reader, yaml.scanner.Scanner, yaml.parser.Parser, yaml.composer.Composer, InstantiableConstructor, yaml.resolver.Resolver):
    def __init__(
        self,
        stream,
        avoidance = [os],
        args_id = special_keys.args_id,
        kwargs_id = special_keys.kwargs_id,
        run_id = special_keys.run_id,
        import_id = special_keys.import_id,
        imported_targets_id = special_keys.imported_targets_id
    ):
        yaml.reader.Reader.__init__(self, stream)
        yaml.scanner.Scanner.__init__(self)
        yaml.parser.Parser.__init__(self)
        yaml.composer.Composer.__init__(self)
        InstantiableConstructor.__init__(self, avoidance, args_id, kwargs_id, run_id, import_id, imported_targets_id)
        yaml.resolver.Resolver.__init__(self)


class Dictspace(dict):
    """
    Simple object for storing attributes.

    Implements equality by attribute names and values, and provides a simple
    string representation.
    """
    def __getitem__(self, name):
        return self.get(name)
    
    def __getattr__(self, name):
        return self[name]

    def __eq__(self, other):
        if not isinstance(other, Dictspace):
            return NotImplemented
        return vars(self) == vars(other)

    def __contains__(self, key):
        return key in self


class Namespace(Dictspace):
    def __repr__(self):
        arg_strings = []
        star_args = {}
        for arg in self.__get_args():
            arg_strings.append(repr(arg))
        for name, value in self.__get_kwargs():
            if name.isidentifier():
                arg_strings.append('%s=%r' % (name, value))
            else:
                star_args[name] = value
        if star_args:
            arg_strings.append('**%s' % repr(star_args))
        return '{%s}' % (', '.join(arg_strings))

    def __get_kwargs(self):
        return list(self.items())

    def __get_args(self):
        return []


from .functions import execute_constructor, load
from .executable_construction import ExecutableConstructor, ExecutableLoader
from .instantiable_construction import InstantiableConstructor, InstantiableLoader
from . import special_keys
from . import namespace


import collections.abc
import os
import importlib
from typing import Optional, Dict, Sequence
import yaml
import yaml.composer
import yaml.parser
import yaml.scanner
from . import special_keys
from .functions import get_import_renames, flatten_list
from . import construction
from .utils.restricted_parse import avoidance_eval

def exec_locals(path: str):
    '''
    Return the local variables after executaing any commands.
    '''
    exec(path)
    return locals()

class ExecutableConstructor(yaml.constructor.SafeConstructor):
    '''
    Distinguish whether the current value is a string or an executable Python code
    by whether there is wrapped by `""` or `''`. If the code is executable, the
    return value will be assigned to the current key. In addition, the key import and
    key run functions can also be used, which will not store the execution results.
    
    Moreover, modules or classes can be renamed by simply using '(name_1, name_2, ...)'
    at the end of the key import, which means that the imported packages or classes can
    be accessed with a new name in another location.
    '''
    def __init__(
        self,
        avoidance = [os],
        run_id = special_keys.run_id,
        import_id = special_keys.import_id,
        imported_targets_id = special_keys.imported_targets_id
    ):
        super().__init__()
        self.avoidance = avoidance
        self.__run_id__ = run_id
        self.__import_id__ = import_id
        self.__package__: Optional[str] = None
        self.__vars__ = {}
        self.__import__ = {}
        self.__imported_targets_id__ = imported_targets_id
    
    def construct_yaml_str(self, node):
        return construction.construct_yaml_str(self, node, None, self.__vars__, avoidance=self.avoidance)
    
    def construct_yaml_seq(self, node):
        return construction.construct_yaml_seq(self, node)
    
    def import_target(self, path: str, returns_dict=False):
        last_point_index = path.rfind('.')
        if last_point_index == -1:
            while True:
                try:
                    name = path
                    target = avoidance_eval(name, None, self.__vars__, avoidance=self.avoidance)
                    final_dict = [{name : target}]
                    break
                except NameError:
                    self.__vars__.update(exec_locals(f'import {name}'))
                except:
                    raise yaml.constructor.ConstructorError(f"No such module or member named as `{name}`")
        else:
            target = path[last_point_index + 1:]
            if target == '*':
                final_dict = exec_locals(f'from {path[:last_point_index]} import *')
            else:
                module = importlib.import_module(path[:last_point_index], self.__package__)
                final_dict = []
                targets = target.split(',')
                for target in targets:
                    target = target.strip()
                    name = target
                    target = getattr(module, target)
                    final_dict.append({name : target})
        if returns_dict:
            return final_dict
        else:
            return target
    
    def escape_special_key(self, key, special_key):
        return key
    
    def check_key_format(self, key, key_node):
        pass
    
    def construct_value(self, parent_node, node, deep=False):
        return self.construct_object(node, deep)
    
    def import_target_from_node(self, node, key_node, value_node, deep=False):
        key = self.construct_object(key_node, deep=deep)
        if not isinstance(key, collections.abc.Hashable):
            raise yaml.constructor.ConstructorError("while constructing a mapping", node.start_mark,
                    "found unhashable key", key_node.start_mark)
        import_id_length = len(self.__import_id__)
        if key[:import_id_length] == self.__import_id__:
            if value_node.style is not None:
                raise yaml.constructor.ConstructorError("The import value should not be a `str`, which is wrapped in `\"\"` or `''`.", value_node.start_mark)
            path = self.construct_scalar(value_node)
            final_dict = self.import_target(path, True)
            vars_dict = {}
            if len(key) > import_id_length:
                if isinstance(final_dict, Dict):
                    raise yaml.constructor.ConstructorError(
                        "Renames of imported targets cannot be used when importing all from a module.",
                        key_node.start_mark)
                as_names = get_import_renames(key[import_id_length:].strip())
                assert len(as_names) == len(final_dict), 'Renames are not matched with imported targets.'
                for as_name, pair in zip(as_names, final_dict):
                    if isinstance(as_name, Sequence):
                        as_name = flatten_list(as_name)
                        for name in as_name:
                            for _, target in pair.items():
                                vars_dict.update({name : target})
                    else:
                        for _, target in pair.items():
                            vars_dict.update({as_name : target})
            else:
                if isinstance(final_dict, Dict):
                    vars_dict = final_dict
                else:
                    for pair in final_dict:
                        for name, target in pair.items():
                            vars_dict.update({name : target})
            self.__import__.update(vars_dict)
            self.__vars__.update(vars_dict)
        else:
            key = self.escape_special_key(key, self.__import_id__)
            self.check_key_format(key, key_node)
            value = self.construct_value(node, value_node, deep=deep)
            return key, value
    
    def construct_mapping(self, node, deep=False):
        if isinstance(node, yaml.MappingNode):
            self.flatten_mapping(node)
        else:
            raise yaml.constructor.ConstructorError(None, None,
                    "expected a mapping node, but found %s" % node.id,
                    node.start_mark)
        mapping = {}
        for key_node, value_node in node.value:
            pair = self.import_target_from_node(node, key_node, value_node, deep)
            if pair is not None:
                key, value = pair
                if key != self.__run_id__:
                    mapping[key] = value
        return mapping

ExecutableConstructor.add_constructor('tag:yaml.org,2002:str', ExecutableConstructor.construct_yaml_str)
ExecutableConstructor.add_constructor('tag:yaml.org,2002:seq', ExecutableConstructor.construct_yaml_seq)

class ExecutableLoader(yaml.reader.Reader, yaml.scanner.Scanner, yaml.parser.Parser, yaml.composer.Composer, ExecutableConstructor, yaml.resolver.Resolver):
    def __init__(
        self,
        stream,
        avoidance = [os],
        run_id = special_keys.run_id,
        import_id = special_keys.import_id,
        imported_targets_id = special_keys.imported_targets_id
    ):
        yaml.reader.Reader.__init__(self, stream)
        yaml.scanner.Scanner.__init__(self)
        yaml.parser.Parser.__init__(self)
        yaml.composer.Composer.__init__(self)
        ExecutableConstructor.__init__(self, avoidance, run_id, import_id, imported_targets_id)
        yaml.resolver.Resolver.__init__(self)


from lark import Transformer, Lark

parser = Lark(
r"""
?value: list | name

!name: NAME
NAME: /[^\W\d]\w*/
list : "(" [value ("," value)*] ")"

%import common.WS
%ignore WS
""", start='value')

class YAMLImportRenamesTransformer(Transformer):
    def name(self, value):
        return str(value[0])
    list = list

def get_import_renames(text):
    tree = parser.parse(text)
    return YAMLImportRenamesTransformer().transform(tree)


from typing import Iterable, Any
import os
import yaml
from .construction import construct_yaml_str, construct_yaml_seq, back_locals
from .utils.yaml_import_renames import get_import_renames
from .namespace import Dictspace
from . import special_keys

def load(
    stream,
    is_instantiable: bool = True,
    avoidance: Iterable[Any] = [os],
    *,
    run_id = special_keys.run_id,
    import_id = special_keys.import_id,
    imported_targets_id = special_keys.imported_targets_id,
    args_id = special_keys.args_id,
    kwargs_id = special_keys.kwargs_id,
):
    """
    Parse the first YAML document in a stream
    and produce the corresponding Python object.
    """
    if is_instantiable:
        from .instantiable_construction import InstantiableLoader
        loader = InstantiableLoader(stream, avoidance, args_id, kwargs_id, run_id, import_id, imported_targets_id)
    else:
        from .executable_construction import ExecutableLoader
        loader = ExecutableLoader(stream, avoidance, run_id, import_id, imported_targets_id)
    loader.__package__ = __package__
    loader.__vars__.update(back_locals())
    try:
        data = loader.get_single_data()
        if len(loader.__import__) > 0:
            setattr(data, loader.__imported_targets_id__, Dictspace(loader.__import__))
        return data
    finally:
        loader.dispose()


def execute_constructor(constructor: yaml.constructor.Constructor):
    constructor.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)
    constructor.add_constructor('tag:yaml.org,2002:seq', construct_yaml_seq)


def flatten_list(nested_list):
    stack = [nested_list]
    flat_list = []
    while stack:
        current = stack.pop()
        if isinstance(current, list):
            for item in reversed(current):
                stack.append(item)
        else:
            flat_list.append(current)
    return flat_list


'''
This module can help users quickly modify the default special
key identifiers.
'''

import_id = '__import__'
imported_targets_id = '__import__'
args_id = '__args__'
kwargs_id = '__kwargs__'
run_id = '__run__'